from primavera import PrimerSelector, Primer, load_record
import os

# LOAD ALL RECORDS TO ANALYSE AND AVAILABLE PRIMERS
constructs_path = os.path.join('data', 'constructs')
records = [
    load_record(os.path.join(constructs_path, f), linear=False)
    for f in sorted(os.listdir(constructs_path)) if f.endswith('.gb')
]
primers_path = os.path.join('data', "available_primers.fa")
available_primers = Primer.list_from_fasta(primers_path)


# SELECT THE BEST PRIMERS
selector = PrimerSelector(read_range=(150, 800), tm_range=(55, 70))
selected_primers = selector.select_primers(records, available_primers)


# PLOT THE PREDICTED SEQUENCING COVERAGE FOR EACH CONSTRUCT
selector.plot_coverage(records=records, selected_primers=selected_primers,
                       pdf_path='primer_selection_example.pdf')


# WRITE ALL PRIMERS IN A CSV FILE (PRIMERS TO ORDER ARE FIRST)
selector.write_primers_table(selected_primers=selected_primers,
                             csv_path='primer_selection_example.csv')



""" dna_sequencing_viewer/__init__.py """

# __all__ = []

from .ReadReferenceMatches import (ReadReferenceMatches, SequencingRead,
                                   ReadReferenceMatchesSet)
from .Primer import Primer
from .SequencingReportGenerator import (SequencingReportGenerator,
                                        PrimersFastaSource,
                                        ConstructsFolderSource)
from .sequencing_simulation import simulate_sequencing
from .PrimerSelector import PrimerSelector
from .biotools import annotate_record, blast_sequences, load_record
from .version import __version__



"""Class for one-line sequencing report generation from a sequencing batch"""

import os
from collections import defaultdict

import matplotlib.pyplot as plt
from Bio import SeqIO
from flametree import file_tree
from .ReadReferenceMatches import SequencingRead, ReadReferenceMatches
from .Primer import Primer


class Source:

    def __call__(self, name):
        return self.get(self.sanitize_name(name), None)

    @staticmethod
    def sanitize_name(name):
        if "." in name:
            name = name.split(".")[0]
        return name.lower().replace(" ", "_").replace("-", "_")


class PrimersFastaSource(Source):
    """Primer source using a fasta file for primers names and sequences."""

    def __init__(self, fasta_file):
        self.fasta_file = fasta_file
        self.primers_dict = {
            self.sanitize_name(primer.name): primer
            for primer in Primer.list_from_fasta(fasta_file)
        }
        self.get = self.primers_dict.get

class ConstructsFolderSource(Source):
    """Loads all files in the folders as Biopython records.
    """

    def __init__(self, dirs, extensions=("gb", "gbk"), file_format="genbank"):
        self.records = {}
        for c_folder in dirs:
            self.records.update({
                self.sanitize_name(name):
                    SeqIO.read(os.path.join(c_folder, name), file_format)
                for name in os.listdir(c_folder)
                if os.path.splitext(name)[1][1:] in extensions
            })
        self.get = self.records.get


class SequencingReportGenerator:
    """Reads a series of sequencing reads from a zip. Make a report"""

    def __init__(self, primers_source=None, constructs_source=None,
                 default_linearity=False):
        self.primers_source = primers_source
        self.constructs_source = constructs_source
        self.default_linearity = default_linearity

    def get_primer_and_construct_names(self, filename):
        """Assume sequencing filenames of the form construct_primer_xxx.ab1"""
        return filename.split("_")[:2]

    def classify_reads(self, reads):
        """Return a dict {construct_id: [associated_reads]}."""
        classified_reads = defaultdict(lambda *a: [])
        for read in reads:
            primer_name, construct_name = \
                self.get_primer_and_construct_names(read.read_name)
            read.primer = self.primers_source(primer_name)

            classified_reads[construct_name].append(read)
        for construct, seqlist in classified_reads.items():
            seqlist.sort(key=lambda r: r.primer.name)
        self.classified_reads = classified_reads
        return classified_reads

    def plot_matches_set(self, matches_set, filepath, title=None):
        ax = matches_set.plot(
            plot_coverage=True,
            plot_reference=True, reference_ax=None,
            figsize="auto", features_filters=(),
            features_properties=None, reference_reads_shares="auto")
        if title is not None:
            ax.set_title(title)
        ax.figure.savefig(filepath, format="png", bbox_inches="tight")
        plt.close(ax.figure)

    def make_report(self, ab1_files=None, ab1_zip_file=None,
                    target="my_folder", perc_identity=100,
                    plot_params=None, replace=True):
        """"""
        if ab1_zip_file is not None:
            ab1_files = [
                f for f in file_tree(ab1_files)._all_files
                if f._extension == "ab1"
            ]
        reads = [SequencingRead.from_ab1_file(f) for f in ab1_files]
        reads = [r for r in reads if set(r.read_sequence) != set('N')]

        classified_reads = self.classify_reads(reads)
        errors = []
        records = []

        root = file_tree(target, replace=replace)

        for construct_name, reads in classified_reads.items():
            construct = self.constructs_source(construct_name)
            if construct is None:
                errors.append(construct_name + " unknown")
                continue
            linear = construct.__dict__.get('linear', self.default_linearity)
            matches_set = ReadReferenceMatchesSet.from_reads(
                construct, reads, perc_identity=perc_identity, linear=linear
            )
            self.plot_matches_set(matches_set, title=construct_name,
                                  filepath=root._file(construct_name + ".png"))
            matches_set.to_genbank(root._file(construct_name + ".gb"))
            records += [
                dict(
                    read=read_name,
                    construct=construct_name,
                    primer=matches.primer.name,
                    primer_position=matches.primer_start,
                    farthest_reading_span=matches.farthest_reading_span,
                    longest_match=matches.longest_match_size,
                    total_matches_length=matches.total_matches_length,
                    average_read_quality=matches.read.average_quality,
                    read_length=len(matches.read.read_qualities)
                )
                for read_name, matches in
                sorted(matches_set.sequencing_matches.items())
            ]
        columns = ["read", "construct", "primer", "primer_position",
                   "longest_match", "total_matches_length",
                   "average_read_quality", "read_length"]
        csv_content = "\n".join(
            [",".join(columns)] + [
                ",".join([('' if (record[col] is None) else str(record[col]))
                          for col in columns])
                for record in records
            ]
        )
        root._file("report.csv").write(csv_content)
        return root._close(), errors



import numpy as np

def minimal_cover(elements_set, subsets, max_subsets=None, heuristic='default',
                  selected=(), depth=0):
    """

    Parameters
    ----------
    elements_set
      The set of all ements to cover

    subsets
      A list of (name, subset)

    max_subsets
      Maximal number of subsets allowed

    heuristic
      A function ``((name, subset), selected) => value`` where ``name`` is the
      name of a subset, ``subset`` is what remains of the subset at this stage,
      ``selected`` is a list of already-selected subset names.

    selected
      (Recursion parameter, do not use.) Already-selected elements

    depth
      (Recursion parameter, do not use.). Depth of the recursion

    Returns
    -------

      None if no solution was found, else a collection of [(name, subset)...]
      in the order in which the subsets
    """

    if len(elements_set) == 0:
        return []
    if max_subsets == 0:
        return None

    if depth == 0:
        full_set = set().union(*[subset for name, subset in subsets])
        if full_set != elements_set:
            return None

    subsets = [(n, s) for (n, s) in subsets if len(s)]

    def sorting_heuristic(named_subset):
        name, subset = named_subset
        if (heuristic == 'default'):
            return len(subset)
        else:
            return heuristic(named_subset, selected)

    ordered_subsets = sorted(subsets, key=sorting_heuristic)

    while len(ordered_subsets):
        if max_subsets is not None:
            critical_subset_length = len(elements_set) / max_subsets
            max_len = max(len(s) for name, s in ordered_subsets)
            if max_len < critical_subset_length:
                return None
        name, subset = ordered_subsets.pop()
        new_elements_set = elements_set.difference(subset)
        new_subsets = [
            (name_, sub.difference(subset))
            for (name_, sub) in ordered_subsets
        ]
        new_max_subsets = None if (max_subsets is None) else max_subsets - 1
        result = minimal_cover(new_elements_set, new_subsets,
                               heuristic=heuristic,
                               selected=list(selected) + [subset],
                               max_subsets=new_max_subsets,
                               depth=depth + 1)
        if result is not None:
            return result + [name]
        ordered_subsets = [
            subset_
            for (subset_, (new_name, new_subset)) in zip(ordered_subsets,
                                                         new_subsets)
            if len(new_subset) != 0
        ]
    return None

def segments_to_array(segments, array_length):
    array = np.zeros(array_length)
    for start, end in segments:
        array[start: end] = 1
    return array

def group_overlapping_segments(segments, min_distance=10):
    if segments == []:
        return []
    returned_segments = [list(segments[0])]
    for start, end in segments[1:]:
        if start < returned_segments[-1][-1] + min_distance:
            if end > returned_segments[-1][-1]:
                returned_segments[-1][-1] = end
        else:
            returned_segments.append([start, end])
    return [tuple(s) for s in returned_segments]



"""All things sequencing simulation."""

import re
from collections import OrderedDict
from .ReadReferenceMatches import (SequenceMatch, ReadReferenceMatches,
                                   ReadReferenceMatchesSet)
from .biotools import reverse_complement

def simulate_sequencing(sequence, primers, read_range=(100, 800),
                        linear=True, trim_nonmatching_primers=True):
    """

    Parameters
    ----------

    sequence

    primers

    read_length

    gap

    linear

    trim_nonmatching_primers

    Returns
    --------

    If primers is simply a Primer object, a SequencingMatches object
    featuring the primer matches on the sequence (`match.primer_matches`)
    and the associated read(s) (`match.read_matches`)

    If `primers` is an list/tuple of primers, an ordered dictionnary of the
    form {primer_name: SequencingMatches()}
    """

    if isinstance(primers, (list, tuple)):
        result = OrderedDict([
            (p.name, simulate_sequencing(sequence=sequence, primers=p,
                                         read_range=read_range, linear=linear))
            for p in primers
        ])
        if trim_nonmatching_primers:
            for name, matches in list(result.items()):
                if matches.read_matches == []:
                    result.pop(name)
        if len(result) == 0:
            return None
        return ReadReferenceMatchesSet(result, linear=linear)

    primer = primers

    len_seq = len(sequence)
    record = sequence
    if not linear:
        sequence = sequence + sequence

    if hasattr(sequence, "seq"):
        sequence = str(record.seq)

    def simulate_one_read(reverse):

        seq = reverse_complement(sequence) if reverse else sequence

        primer_matches = [
            SequenceMatch(start=m.start(), end=m.end())
            for m in re.finditer(primer.sequence, seq)
        ]
        read_min, read_max = read_range
        read_matches = [
            SequenceMatch(start=m.start + read_min, end=m.start + read_max)
            for m in primer_matches
        ]

        for match_group in (primer_matches, read_matches):
            for match in match_group:
                start, end = match.start, match.end
                if reverse:
                    start, end = len(seq) - end, len(seq) - start
                    match.strand = -1
                if start > len_seq:
                    if linear:
                        match_group.remove(match)
                    else:
                        start, end = start - len_seq, end - len_seq
                elif end < 0:
                    if linear:
                        match_group.remove(match)
                    else:
                        start, end = start + len_seq, end + len_seq
                match.start, match.end = start, end

        return primer_matches, read_matches

    primer_matches, read_matches = simulate_one_read(reverse=False)
    primer_matches_r, read_matches_r = simulate_one_read(reverse=True)
    all_primer_matches = primer_matches + primer_matches_r
    all_read_matches = read_matches + read_matches_r
    return ReadReferenceMatches(record, all_primer_matches, all_read_matches,
                                primer=primer)



"""Bla."""
import tempfile
import subprocess
import time
import os
import sys
import zipfile
from copy import deepcopy

import numpy as np
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation

PYTHON3 = (sys.version_info[0] == 3)

if PYTHON3:
    from io import StringIO, BytesIO
    StringBytesIO = BytesIO

else:
    from StringIO import StringIO
    BytesIO = StringIO
    StringBytesIO = StringIO

def complement(dna_sequence):
    """Return the complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"TACGGC"``.
    Uses BioPython for speed.
    """
    return str(Seq(dna_sequence).complement())


def reverse_complement(sequence):
    """Return the reverse-complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"GCCGTA"``.
    Uses BioPython for speed.
    """
    return complement(sequence)[::-1]


def blast_sequences(sequences=None, fasta_file=None,
                    blast_db=None, subject=None, word_size=4,
                    perc_identity=80, num_alignments=1000, num_threads=3,
                    use_megablast=True, ungapped=True):
    """Return a Biopython BLAST record of the given sequence BLASTed
    against the provided database.

    Parameters
    ----------

    sequences
      Either an ATGC string or a list of ATGC strings or a dict {name: seq:}

    subject
      Either a path to a fasta (.fa) file or an ATGC string. Subject to blast
      against.

    word_size
      Word size to use in the blast

    perc_identity
      Minimal percentage of identical nucleotides in a match for it to be kept

    num_alignments
      Number of alignments to keep

    num_threads
      Number of threads for the BLAST

    use_megablast
      Whether to use Megablast.

    ungapped
      No-gaps matches only ?



    Examples
    --------

    >>> blast_record = blast_sequence("ATTGTGCGTGTGTGCGT", "blastdb/ecoli")
    >>> for alignment in blast_record.alignments:
    >>>     for hit in alignment.hsps:
    >>>         print (hit.identities)
    """

    if isinstance(sequences, str):
        sequences = [sequences]

    if isinstance(sequences, (list, tuple)):
        sequences = {"seq_%d" % i: seq for i, seq in enumerate(sequences)}

    xml_file, xml_name = tempfile.mkstemp(".xml")
    fasta_file, fasta_name = tempfile.mkstemp(".fa")
    with open(fasta_name, "w+") as f:
        for (name, seq) in sequences.items():
            f.write(">%s\n%s\n" % (name, seq))

    remove_subject = True
    if subject is not None:
        close_subject = True
        if isinstance(subject, str):
            if subject.endswith(".fa"):
                remove_subject = False
            else:
                subject = [subject]
        if isinstance(subject, (list, tuple)):
            subject = {"subject_%d" % i: seq for i, seq in enumerate(subject)}
        if isinstance(subject, dict):
            subject_file, fasta_subject_name = tempfile.mkstemp(".fa")
            with open(fasta_subject_name, "w+") as f:
                for (name, seq) in subject.items():
                    f.write(">%s\n%s\n" % (name, seq))
            subject = fasta_subject_name
    else:
        close_subject = False

    p = subprocess.Popen([
        "blastn", "-out", xml_name,
        "-outfmt", "5",
        "-num_alignments", str(num_alignments),
        "-query", fasta_name] +
        (["-db", blast_db] if blast_db is not None
         else ['-subject', subject]) +
        (["-ungapped"] if ungapped else []) +
        (["-task", "megablast"] if use_megablast else []) + [
        "-word_size", str(word_size),
        "-num_threads", str(num_threads),
        "-perc_identity", str(perc_identity)
    ], close_fds=True, stderr=subprocess.PIPE)
    res, blast_err = p.communicate()
    p.wait()
    error = None
    for i in range(3):
        try:
            with open(xml_name, "r") as f:
                res = list(NCBIXML.parse(f))
                os.fdopen(xml_file, 'w').close()
                os.fdopen(fasta_file, 'w').close()
                os.remove(xml_name)
                os.remove(fasta_name)

                if close_subject:
                    open(subject, 'w').close()
                    if remove_subject:
                        os.remove(subject)
                return res
            break
        except ValueError as err:
            error = err
            time.sleep(0.1)
    else:
        raise ValueError("Problem reading the blast record: " + str(error))


# RETIRING THIS ONE ASAP:
def read_records_from_zip(zip_path):
    """Return SeqRecords from all FASTA/GENBANK files in the zip."""
    with zipfile.ZipFile(zip_path, 'r') as archive:
        extensions_types = {".ab1": "abi", ".abi": "abi", ".gb": "genbank",
                            ".gbk": "genbank", ".fa": "fasta",
                            ".fasta": "fasta"}
        extract = {}
        failed_files = []
        for f in archive.filelist:
            name, ext = os.path.splitext(f.filename)
            try:
                if ext in extensions_types:
                    content = StringBytesIO(archive.read(f.filename))
                    extract[f.filename] = SeqIO.read(content,
                                                     extensions_types[ext])
            except:
                failed_files.append(f.filename)
    return extract, failed_files

def rotate_circular_record(record, n_bases):
    """Changes the starting point of a circular SeqRecord by n_bases bases."""
    new_record = deepcopy(record)
    new_record.seq = record.seq[n_bases:] + record.seq[:n_bases]
    for f in new_record.features:
        f.location += (-n_bases)
        if max(f.location.start, f.location.end) <= 0:
            f.location += len(record)
    return new_record

def group_overlapping_segments(segments, min_distance=10):
    if segments == []:
        return []
    returned_segments = [list(segments[0])]
    for start, end in segments[1:]:
        if start < returned_segments[-1][-1] + min_distance:
            if end > returned_segments[-1][-1]:
                returned_segments[-1][-1] = end
        else:
            returned_segments.append([start, end])
    return [tuple(s) for s in returned_segments]



def get_segment_coordinates(center, segment_length, sequence_length):
    """Return max(0, c - s/2) - min(L, c + L/2).

    Where c=center, s=segment_length, L=sequence_length.
    """
    half = int(segment_length / 2)
    start = max(0, min(center - half,  sequence_length - segment_length))
    end = start + segment_length
    return start, end

def find_best_primer_locations(sequence, size_range=(15, 25),
                               tm_range=(55, 70)):
    """Quickly compute all overhangs in the sequence.

    This function uses the heuristic {A, T}=2degC, {G, C}=4degC to compute
    melting temperatures.

    This function uses vectorial operations for speed. The results are also
    cached.
    """
    lmin, lmax = size_range
    tmin, tmax = tm_range

    table = np.zeros((lmax + 1 - lmin, len(sequence)))
    cumsum = np.cumsum([4 if nuc in "GC" else 2 for nuc in sequence])
    for i, oh_size in enumerate(range(lmin, lmax + 1)):
        arr = cumsum[oh_size:] - cumsum[:-oh_size]
        start = int(oh_size / 2)
        end = start + len(arr)
        table[i, start:end] = arr
        table[i, :start] = table[i, start]
        table[i, end:] = table[i, end-1]
    scores = - (table - tmin) * (table - tmax)
    best_sizes_indices = scores.argmax(axis=0)
    best_sizes = lmin + best_sizes_indices
    validities = np.choose(best_sizes_indices, scores) >= 0
    osizes_and_validities = zip(best_sizes, validities)
    return [
      None if not valid
      else get_segment_coordinates(i, ovh_size, len(sequence))
      for i, (ovh_size, valid) in enumerate(osizes_and_validities)
    ]

def find_non_unique_segments(sequence, perc_identity=80):
    blast_record = blast_sequences(sequence, subject=sequence,
                                   perc_identity=perc_identity,
                                   ungapped=False, word_size=7)[0]
    segments_with_alignments = sorted(set([
        (h.query_start, h.query_end)
        for al in blast_record.alignments
        for h in al.hsps
        if (h.query_start, h.query_end) != (1, len(sequence))
    ]))
    return group_overlapping_segments(segments_with_alignments)

def load_record(filename, linear=True, name='auto'):
    if filename.lower().endswith(("gb", "gbk")):
        record = SeqIO.read(filename, "genbank")
    elif filename.lower().endswith(('fa', 'fasta')):
        record = SeqIO.read(filename, "fasta")
    else:
        raise ValueError('Unknown format for file: %s' % filename)
    record.linear = linear
    if name == 'auto':
        name = os.path.splitext(os.path.basename(filename))[0]
    record.id = name
    record.name = name.replace(" ", "_")[:20]
    return record

def annotate_record(seqrecord, location="full", feature_type="misc_feature",
                    margin=0, **qualifiers):
    """Add a feature to a Biopython SeqRecord.

    Parameters
    ----------

    seqrecord
      The biopython seqrecord to be annotated.

    location
      Either (start, end) or (start, end, strand). (strand defaults to +1)

    feature_type
      The type associated with the feature

    margin
      Number of extra bases added on each side of the given location.

    qualifiers
      Dictionnary that will be the Biopython feature's `qualifiers` attribute.
    """
    if location == "full":
        location = (margin, len(seqrecord) - margin)

    strand = location[2] if len(location) == 3 else 1
    seqrecord.features.append(
        SeqFeature(
            FeatureLocation(location[0], location[1], strand),
            qualifiers=qualifiers,
            type=feature_type
        )
    )



"""Functions and Classes for matching sequencing results to a reference."""

from copy import deepcopy
import matplotlib.pyplot as plt
from dna_features_viewer import (BiopythonTranslator, GraphicRecord,
                                 GraphicFeature)
import numpy as np
from .biotools import (blast_sequences, read_records_from_zip,
                       rotate_circular_record)
from collections import OrderedDict
import matplotlib.gridspec as gridspec
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
import os


class SequencingRead:
    """Class representing a (Sanger) sequencing read.

    It is recommended to build SequencingReads using the `.from_ab1_file`
    method.

    Parameters
    ----------

    read_name
      Name (label or any other ID) of the read

    read_sequence
      An ATGC string


    """

    def __init__(self, read_name=None, read_sequence=None, read_qualities=None,
                 primer=None):
        self.read_name = read_name
        self.primer = primer
        self.read_sequence = read_sequence
        self.read_qualities = read_qualities
        self.average_quality = (None if read_qualities is None else
                                np.mean(read_qualities))

    def __hash__(self):
        return self.name

    @staticmethod
    def from_ab1_file(ab1_file, read_name=None, primer=None):
        """Read a read from Ab1 data.

        ab1_file can be a file path, a file handle, or a Flametree file object.
        """
        if hasattr(ab1_file, "_name_no_extension"): # Flametree
            read_name = ab1_file._name_no_extension
            ab1_file = ab1_file.open('rb')

        if read_name is None:
            read_name = os.path.basename(os.path.splitext(ab1_file)[0])
        record = SeqIO.read(ab1_file, "abi")
        return SequencingRead(
            read_name=read_name,
            primer=primer,
            read_sequence=str(record.seq),
            read_qualities=np.array(record.letter_annotations["phred_quality"])
        )

    @staticmethod
    def list_from_ab1_zipfile(zipfile):
        reads, errors = read_records_from_zip(zipfile)
        return [
            SequencingRead(
                read_name=name,
                read_sequence=str(record.seq),
                read_qualities=np.array(
                    record.letter_annotations["phred_quality"]))
            for (name, record) in sorted(reads.items())
        ], errors



class SequenceMatch:

    def __init__(self, start, end, strand=1, percent=None,
                 read_qualities=None):
        self.start = start
        self.end = end
        self.strand = strand
        self.percent = percent
        self.read_qualities = read_qualities

    @staticmethod
    def from_hsp(hsp, read=None):
        prc = ("%.01f" % (100.0 * hsp.identities / hsp.align_length)) + "%"
        start, end, strand = hsp.sbjct_start, hsp.sbjct_end, 1
        if start > end:
            start, end, strand = end, start, -1
        if read is not None:
            qualities = read.read_qualities[hsp.query_start:hsp.query_end]
        else:
            qualities = None

        return SequenceMatch(start=start, end=end, strand=strand, percent=prc,
                             read_qualities=qualities)

    def rotated(self, n_bases, construct_length):
        if (self.end - n_bases <= 0):
            n_bases = n_bases - construct_length
        elif (self.start - n_bases >= construct_length):
            n_bases = n_bases + construct_length
        return (self + (-n_bases))

    def __add__(self, number):
        new_match = deepcopy(self)
        new_match.start += number
        new_match.end += number
        return new_match

    def __len__(self):
        return self.end - self.start

    def __repr__(self):
        return "Match(%d, %d, %d, %s)" % (self.start, self.end, self.strand,
                                          self.percent)

    def to_Biopython_feature(self):
        return SeqFeature(
            location=FeatureLocation(self.start, self.end, self.strand),
            type="misc_feature",
            qualifiers={"label": self.percent}
        )


class ReadReferenceMatches:
    """Represent all matches between one read from one primer, and a reference.
    """

    def __init__(self, reference, primer_matches=(), read_matches=(),
                 primer=None, read=None):
        self.primer_matches = list(primer_matches)
        self.read_matches = list(read_matches)
        self.reference = reference
        self.primer = primer
        self.read = read

    def remove_read_matches_contained_in_others(self):
        """Remove every read match contained in other matches."""
        matches = sorted(self.read_matches, key=lambda m: m.start)[::-1]
        new_matches = [m for m in matches]
        for i1, m1 in enumerate(matches):
            for m2 in matches[i1 + 1:]:
                if ((m2.start <= m1.start <= m2.end) and
                    (m2.start <= m1.end <= m2.end)):
                    new_matches.remove(m1)
                    break
        self.read_matches = new_matches

    def rotated(self, n_bases):
        if isinstance(self.reference, str):
            rotated_ref = self.reference[n_bases:] + self.reference[n_bases]
        else:
            rotated_ref = rotate_circular_record(self.reference, n_bases)
        return ReadReferenceMatches(
            rotated_ref,
            primer_matches=[
                m.rotated(n_bases, len(self.reference))
                for m in self.primer_matches
            ],
            read_matches=[
                m.rotated(n_bases, len(self.reference))
                for m in self.read_matches
            ]
        )

    @property
    def primer_gap(self):
        if len(self.read_matches) == 0:
            raise ValueError("Gap computing requires non-empty reads")

        if len(self.primer_matches) != 1:
            return "NA (%d primer matches)" % len(self.primer_matches)
        primer_match = self.primer_matches[0]
        if primer_match.strand == -1:
            read_start = max([m.end for m in self.read_matches])
            primer_start = primer_match.end
            gap = primer_start - read_start
        else:
            read_start = min([m.start for m in self.read_matches])
            primer_start = primer_match.start
            gap = read_start - primer_start
        if gap < 0:
            gap += len(self.reference)
        return gap

    @property
    def farthest_reading_span(self):
        """Return the distance between the primer and the furthest match."""
        if len(self.read_matches) == 0:
            return 0

        if len(self.primer_matches) != 1:
            return "NA (%d primer matches)" % len(self.primer_matches)
        primer_match = self.primer_matches[0]
        if primer_match.strand == -1:
            read_end = min([m.start for m in self.read_matches])
            read_start = min([m.start for m in self.read_matches])
            span = read_start - read_end
        else:
            read_end = max([m.end for m in self.read_matches])
            read_start = min([m.start for m in self.read_matches])
            span = read_end - read_start
        if span < 0:
            span += len(self.reference)
        return span

    @property
    def longest_match_size(self):
        if len(self.read_matches) == 0:
            return 0
        return max(len(match) for match in self.read_matches)

    @property
    def total_matches_length(self):
        return sum([len(match) for match in self.read_matches])

    @property
    def primer_start(self):
        if len(self.primer_matches) == 0:
            return None
        if len(self.primer_matches) == 1:
            return self.primer_matches[0].start
        else:
            return [m.start for m in self.primer_matches]

    def matches_as_Biopython_features(self):
        """Return a list of the matches as Biopython features"""
        result = []
        for m in self.read_matches:
            feature = m.to_Biopython_feature()
            if self.primer is not None:
                feature.qualifiers["primer"] = self.primer.name
            result.append(feature)
        for m in self.primer_matches:
            feature = m.to_Biopython_feature()
            if self.primer is not None:
                feature.qualifiers["label"] = self.primer.name
            result.append(feature)
        return result

    def to_SeqRecord(self):
        """Return a Biopython Seqrecord with the matches as annotations.

        The record also features the reference's features.

        """
        record = deepcopy(self.reference)
        record.features += self.matches_as_Biopython_features()
        return record

    def to_genbank(self, filename):
        """Write the Genbank record with the matches as annotations.
        The record also features the reference's features.
        """
        SeqIO.write(self.to_SeqRecord(), filename, "genbank")


class ReadReferenceMatchesSet:
    """All matches between a list of sequencing reads and one reference seq.

    Parameters
    ----------

    read_reference_matches

    linear
      Whether the reference is linear
    """

    def __init__(self, read_reference_matches, linear=True):

        self.read_reference_matches = read_reference_matches
        self.linear = linear
        self.reference = list(read_reference_matches.values())[0].reference
        self.coverage = np.zeros(len(self.reference))
        for matches in read_reference_matches.values():
            for match in matches.read_matches:
                start, end = sorted([match.start, match.end])
                inds = np.arange(start, end) % len(self.coverage)
                self.coverage[inds] += 1

    def rotated(self, n_bases="auto"):
        if n_bases == "auto":
            nz = np.nonzero(np.hstack([self.coverage, self.coverage]))[0]
            if len(nz) > 0:
                i = np.argmax(np.diff(nz))
                x1, x2 = nz[i], nz[i+1]
                n_bases = int((x1 + x2) / 2) % len(self.reference)
            else:
                n_bases = 0
        new_matches = OrderedDict([
            (name, m.rotated(n_bases))
            for name, m in self.read_reference_matches.items()
        ])
        return ReadReferenceMatchesSet(new_matches, linear=self.linear)

    @staticmethod
    def from_reads(reference, reads, perc_identity=98,
                   primer_perc_identity=95, min_match_length=20,
                   max_read_length=1000, linear=True):
        reference_sequence = (reference if isinstance(reference, str) else
                              str(reference.seq))
        reads = deepcopy(reads)
        reads_dict = {read.read_name: read for read in reads}
        seq_length = len(reference_sequence)
        read_reference_matches = OrderedDict([
            (read.read_name, ReadReferenceMatches(
                reference, primer=read.primer, read=read))
            for read in reads
        ])
        if not linear:
            reference_sequence = reference_sequence + reference_sequence

        reads_primer_sequences = {
            read.read_name: read.primer.sequence
            for read in reads
            if (read.primer is not None) and
               (read.primer.sequence is not None)
        }

        if len(reads_primer_sequences):
            primers_blast_record = blast_sequences(
                reads_primer_sequences, subject=reference_sequence,
                perc_identity=primer_perc_identity
            )

            for primer_alignments in primers_blast_record:
                read_name = primer_alignments.query
                name_matches = read_reference_matches[read_name].primer_matches
                read = reads_dict[read_name]
                for aligmnent in primer_alignments.alignments:
                    for hsp in aligmnent.hsps:
                        if hsp.align_length < len(read.primer.sequence):
                            continue
                        if hsp.sbjct_start > seq_length:
                            continue
                        match = SequenceMatch.from_hsp(hsp)
                        name_matches.append(match)

        reads_sequences = {
            read.read_name: read.read_sequence
            for read in reads
            if read.read_sequence is not None
        }
        reads_blast_record = blast_sequences(reads_sequences,
                                             subject=reference_sequence,
                                             perc_identity=perc_identity)

        for reads_alignments in reads_blast_record:

            read_name = reads_alignments.query
            primer_matches = read_reference_matches[read_name].primer_matches

            read = reads_dict[read_name]
            for al in reads_alignments.alignments:
                for hsp in al.hsps:
                    if hsp.align_length < min_match_length:
                        continue
                    start, end = hsp.sbjct_start, hsp.sbjct_end
                    if start > end:
                        start, end = end, start

                    if start > seq_length:
                        continue

                    if len(primer_matches) == 1:
                        # use primer location information to restrict matches
                        m = primer_matches[0]
                        if (m.strand == 1):
                            mstarts = [m.start]
                            if not linear:
                                mstarts.append(m.start-seq_length)
                            if not any([(mstart < start <
                                         mstart + max_read_length)
                                        for mstart in mstarts]):
                                continue
                        else:
                            mstarts = [m.start]
                            if not linear:
                                mstarts.append(m.start + seq_length)
                            if not any([(mstart > end >
                                         mstart - max_read_length)
                                        for mstart in mstarts]):
                                continue

                    match = SequenceMatch.from_hsp(hsp, read=read)

                    read_reference_matches[read_name].read_matches.append(match)

        for sequencing_match in read_reference_matches.values():
            sequencing_match.remove_read_matches_contained_in_others()

        return ReadReferenceMatchesSet(read_reference_matches, linear=linear)

    def plot(self, ax=None, plot_coverage=True,
             plot_reference=False, reference_ax=None,
             figsize="auto", features_filters=(),
             features_properties=None, reference_reads_shares="auto"):
        """Plot the sequencing matches.

        Useful to get a general overview of the sequencing (coverage, mutations
        etc.)

        Parameters
        ----------

        ax
        """

        class AnnotationsGraphicTranslator(BiopythonTranslator):

            def compute_feature_color(self, f):
                return "#f9d277"

            def compute_feature_label(self, f):
                return BiopythonTranslator.compute_feature_label(f)[:20]

            def compute_filtered_features(self, features):
                def is_not_parameter(f):
                    label = "".join(f.qualifiers.get('label', ''))
                    return label not in ('cover', 'no_primer')
                return [f for f in features if is_not_parameter(f)]

        if plot_reference:
            translator = AnnotationsGraphicTranslator(
                features_filters=features_filters,
                features_properties=features_properties)
            grecord = translator.translate_record(self.reference)
            if not self.linear:
                grecord.split_overflowing_features_circularly()

        if figsize == "auto":
            figsize = (12, "auto")
        if figsize[1] == "auto":
            sequencing_ax_height = 2 + 0.35 * len(self.read_reference_matches)
            if not plot_reference:
                figure_height = sequencing_ax_height
            else:
                ref_ax, _ = grecord.plot(with_ruler=False,
                                         figure_width=figsize[0])
                ref_fig_height = ref_ax.figure.get_size_inches()[1]
                figure_height = sequencing_ax_height + ref_fig_height
                if reference_reads_shares == "auto":
                    reference_reads_shares = (int(100*ref_fig_height),
                                              int(100*sequencing_ax_height))
                plt.close(ref_ax.figure)
            figsize = (figsize[0], figure_height)
        elif reference_reads_shares == "auto":
            reference_reads_shares = (1, 2)

        if plot_reference:
            if reference_ax is None:
                gs = gridspec.GridSpec(sum(reference_reads_shares), 1)
                fig = plt.figure(figsize=figsize, facecolor="w")
                reference_ax = fig.add_subplot(gs[:reference_reads_shares[0]])
                ax = fig.add_subplot(gs[reference_reads_shares[0]:])

            grecord.plot(reference_ax, with_ruler=False, annotate_inline=True)
            self.plot(ax=ax, plot_coverage=plot_coverage, plot_reference=False)
            ax.set_xlim(reference_ax.get_xlim())
            return ax

        # so the first read in the list gets displayed on top
        read_reference_matches = OrderedDict([
            item for item in list(self.read_reference_matches.items())[::-1]
        ])
        L = len(self.reference)
        if ax is None:
            fig, ax = plt.subplots(1, figsize=figsize)
        ax.set_xlim(-2, L)
        ax.set_ylim(0, len(read_reference_matches) + 2)
        ax.set_yticks(range(1, len(read_reference_matches) + 1))
        ax.set_yticklabels([name for name in read_reference_matches])
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.yaxis.set_ticks_position('left')
        ax.xaxis.set_ticks_position('bottom')

        gr_record = GraphicRecord(sequence_length=L, features=[])

        for i, (read_name, matches) in enumerate(read_reference_matches.items()):
            y = i+1
            ax.axhline(y, ls=":", lw=0.5, color="#aaaaaa", zorder=-1000)
            if matches.primer.metadata.get('available', False):
                color = '#f7a3f6'
            else:
                color = "#a3c3f7"
            for match in matches.read_matches:
                gr_record.features = [GraphicFeature(start=match.start,
                                                     end=match.end,
                                                     strand=match.strand,
                                                     color=color)]
                gr_record.split_overflowing_features_circularly()
                for feature in gr_record.features:
                    gr_record.plot_feature(ax, feature, y, linewidth=0.2)
            for match in matches.primer_matches:
                feature = GraphicFeature(start=match.start, end=match.end,
                                         strand=match.strand, color="#e85558")
                gr_record.plot_feature(ax, feature, y, linewidth=0.2)

        if plot_coverage:
            ax.fill_between(range(len(self.coverage)), self.coverage,
                            zorder=-2000, alpha=0.2, facecolor="#a3c3f7")
        return ax

    def extract_minimal_cover(self):
        """Return a version of the set with the minimal amount of reads that
        cover the same total region as in this set."""

        coverage_dict = {}
        for primer_name, read_matches in self.read_reference_matches.items():
            if len(read_matches.primer_matches) != 1:
                continue
            primer_match = read_matches.read_matches[0]
            coverage_dict[(primer_match.start, primer_match.end)] = primer_name
        coverages = list(coverage_dict.keys())

        selected = [min(coverages, key=lambda start_end: (start_end[0],
                                                          -start_end[1]))]
        while True:
            overlapping = [
                (start, end) for (start, end) in coverages
                if (start <= selected[-1][1] < end)
            ]
            if overlapping != []:
                new_segment = max(overlapping,
                                  key=lambda start_end: start_end[1])
                selected.append(new_segment)
            else:
                new_components = [
                    (start, end) for (start, end) in coverages
                    if (selected[-1][1] < start)
                ]
                if new_components == []:
                    break
                new_segment = min(new_components,
                                  key=lambda start_end: (start_end[0],
                                                         -start_end[1]))
                selected.append(new_segment)
        selected_primers = set(coverage_dict[s] for s in selected)
        return ReadReferenceMatchesSet(
            {
                k: v
                for k, v in self.read_reference_matches.items()
                if k in selected_primers
            },
            linear=self.linear
        )

    def to_SeqRecord(self):
        """Return a Biopython Seqrecord with the matches as annotations.

        The record also features the reference's features.
        """
        record = deepcopy(self.reference)
        for _, matches in self.read_reference_matches.items():
            record.features += matches.matches_as_Biopython_features()
        return record

    def to_genbank(self, filename):
        """Write the Genbank record with the matches as annotations.
        The record also features the reference's features.
        """
        print (filename)
        record = self.to_SeqRecord()
        record.features = [f for f in record.features if f.location is not None]
        SeqIO.write(record, filename, "genbank")



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open('primavera/version.py').read()) # loads __version__

setup(
    name='primavera',
    version=__version__,
    author='Zulko',
    description='Primer selection + data analysis for DNA assembly validation',
    long_description=open('README.rst').read(),
    license='see LICENSE.txt',
    keywords="DNA assembly sequencing primer design",
    packages= find_packages(exclude='docs'),
    install_requires=["numpy", "Biopython", "proglog", 'flametree', 'pandas',
                      'dna_features_viewer', 'dnachisel'])



from Bio import SeqIO
from .biotools import reverse_complement
import pandas

class Primer:
    """Class representing a sequencing primer

    Parameters
    ----------

    name
      Name (label) of the primer

    sequence
      An ATGC string

    metadata
      A dictionnary {field, value}
    """


    def __init__(self, name=None, sequence=None, metadata=None):
        self.name = name
        self.sequence = sequence
        self.metadata = {} if metadata is None else metadata

    def find_location(self, sequence):
        ind = sequence.find(self.sequence)
        strand = 1
        if ind == -1:
            ind = sequence.find(reverse_complement(self.sequence))
            if ind == -1:
                return None
            else:
                strand = -1
        start, end = ind, ind + len(self.sequence)
        return start, end, strand

    def __repr__(self):
        return "%s-%s" % (self.name, self.sequence)

    @staticmethod
    def list_from_fasta(fasta_file):
        records = SeqIO.parse(fasta_file, 'fasta')
        return [
            Primer(name=rec.name, sequence=str(rec.seq))
            for rec in records
        ]

    @staticmethod
    def list_to_fasta(primers_list, filepath=None):
        fasta = "\n\n".join([
            ">%s\n%s" % (primer.name, primer.sequence)
            for primer in primers_list
        ])
        if filepath is not None:
            with open(filepath, 'w') as f:
                f.write(fasta)
        return fasta

    @staticmethod
    def list_from_spreadsheet(dataframe=None, filepath=None):
        if dataframe is None:
            if filepath.lower().endswith('csv'):
                dataframe = pandas.read_csv(filepath)
            else:
                dataframe = pandas.read_excel(filepath)

        return [
            Primer(name=r['name'], sequence=r['sequence'],
                   metadata=dict((k, v) for (k, v) in r.items()
                                 if k not in ['name', 'sequence']))
            for r in dataframe.to_dict(orient='record')
        ]

    @staticmethod
    def list_to_spreadsheet(primers_list, filepath=None):

        dataframe = pandas.DataFrame.from_records(
            columns=['name', 'sequence'] + sorted(set(
                field
                for primer in primers_list
                for field in primer.metadata
            )),
            data=[
                dict([('name', primer.name), ('sequence', primer.sequence)] +
                     list(primer.metadata.items()))
                for primer in primers_list
            ]
        )

        if filepath is not None:
            if filepath.lower().endswith('csv'):
                dataframe.write_csv(filepath)
            else:
                dataframe.write_excel(filepath)
        return dataframe



"""Class to automatically select available and new primers for sequencing runs.
"""

from copy import deepcopy
import re
from collections import defaultdict

import pandas
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import numpy as np

import flametree
from proglog import TqdmProgressBarLogger, ProgressBarLogger
from dnachisel import (AvoidPattern, repeated_kmers, homopolymer_pattern,
                       DnaOptimizationProblem)

from .sequencing_simulation import simulate_sequencing
from .biotools import (reverse_complement, find_non_unique_segments,
                       find_best_primer_locations)
from .tools import minimal_cover, segments_to_array, group_overlapping_segments
from .Primer import Primer


class PrimerSelectorLogger(TqdmProgressBarLogger):
    """Custom logger class adapted to the logger selector."""
    def __init__(self, bars=('record', 'primer'), notebook='default'):
        ignored_bars = set(('record', 'primer')).difference(bars)
        TqdmProgressBarLogger.__init__(self, bars=bars, notebook=notebook,
                                       ignored_bars=ignored_bars,
                                       min_time_interval=0.2)

class PrimerSelector:
    """A selector to compute the best primers to sequence a set of constructs.

    Examples
    --------

    >>> selector = PrimerSelector()
    >>> selected_primers = selector.select_primers(records, available_primers)
    >>> selector.plot_coverage(records, selected_primers, 'my_report.pdf'

    Parameters
    -----------

    read_range
      The experimentally measured range (start, end) so that, when a primer
      anneals in the sequence at index i, the range ``[i + start, i + end]`` will
      be correctly sequenced.

    size_range
      Size range (min, max) for the size of the primers, in nucleotides.

    tm_range
      Acceptable melting temperature range for the primers (in Celsius), as
      computed using the heuristic A/T=2C, G/C=4C

    primer_conditions
      A list of functions of the form ``primer_sequence => True/False``.
      Primers for which at least one condition returns False will not be
      considered.

    primer_reuse_bonus
      Weight that the availability of the primer should have in the decision
      to select this primer. A higher value of this parameter leads to
      solutions where a higher less new primershave to be ordered, but more
      sequencing reactions have to be done. Set to e.g. 200 to test if there
      exists solutions involving solely already-available primers.

    logger
      Leave to 'bars' for default progress-bar logger, to None for no logger,
      or any Proglog ProgressBarLogger object.

    coverage_resolution
      When the user provides a record with "cover" features to indicate where
      to cover, the coverage points used by the algorithm are the 1-in-N
      nucleotides along the feature region, where N is this parameter.
    """

    def __init__(self, read_range=(150, 800), size_range=(16, 25),
                 tm_range=(55, 70), primer_conditions=(),
                 primer_reuse_bonus=2, logger='bars',
                 coverage_resolution=5):
        self.read_range = read_range
        self.size_range = size_range
        self.tm_range = tm_range
        self.primers_conditions = primer_conditions
        self.coverage_resolution = coverage_resolution
        self.primer_reuse_bonus = 2
        if logger == 'bars':
            logger = PrimerSelectorLogger()
        if logger is None:
            logger = ProgressBarLogger()
        self.logger = logger

    def select_primers(self, records, available_primers=(),
                       new_primers_prefix='P', new_primers_digits=6):
        """Select primers to sequence the given records.

        Parameters
        ----------
        records
          A list of biopython records to sequence. The zones to cover in the
          record should be indicated by a feature of type ``misc_feature`` and
          label ``cover``. The zones where no primers are desired should be
          indicated by a feature of type ``misc_feature`` and label
          ``no_primer``.

        available_primers
          List of Primer objects representing the available primers.

        new_primers_prefix
          Prefix to use for names of the new primers

        new_primers_digits
          The new primers will have names of the form P000435, with a number
          of digits provided by this parameter.

        Returns
        -------
        selected_primers
          A list of lists of primers, one list of primers for each consrtruct.

        """

        available_primers_dict = {p.sequence: p for p in available_primers}
        available_primers_seqs = set([p.sequence for p in available_primers])

        # COMPUTE PRIMERS AND COVERAGES
        indices_to_cover = {}
        primers_coverages = defaultdict(lambda *a: set())
        self.logger(message='Analyzing the records...')
        for record in self.logger.iter_bar(record=records):
            indices_to_cover[record.id] = {
                ind: '%s_%03d' % (record.id, ind)
                for ind in self.compute_indices_to_cover(record)
            }
            coverages = self.compute_all_primers_coverage_on_record(
                record, available_primers=available_primers_seqs,
                indices_to_cover=indices_to_cover[record.id])
            for primer, coverage in coverages.items():
                primers_coverages[primer].update(coverage)

        # FIND GLOBAL MINIMAL COVER
        self.logger(message='Selecting primers...')
        elements_set = set(
            index
            for rec_id, named_indices in indices_to_cover.items()
            for index in named_indices.values()
        )

        def heuristic(named_subset, selected):
            name, subset = named_subset
            primer_is_reused = name in available_primers_seqs
            reuse_bonus = self.primer_reuse_bonus * primer_is_reused
            return len(subset) + reuse_bonus

        subsets = primers_coverages.items()
        primers_cover = minimal_cover(elements_set, subsets=subsets,
                                      heuristic=heuristic)

        # REORGANIZE AND NAME THE SELECTED PRIMERS
        available_primers_names = [p.name for p in available_primers]
        selected_primers = []
        selected_primer_from_seq = {}
        for primer_seq in primers_cover:
            if primer_seq in available_primers_dict:
                name = available_primers_dict[primer_seq].name
                primer = available_primers_dict[primer_seq]
                infos = primer.metadata.get('infos', '')
                primer = Primer(primer.name, primer.sequence,
                                metadata={'available': True, 'infos': infos})
            else:
                name = self.generate_primer_name(
                    available_primers_names=available_primers_names,
                    prefix=new_primers_prefix,
                    n_digits=new_primers_digits
                )

                try:
                    part, index = self.find_subsequence_in_records(
                        sequence=primer_seq, records=records)
                    infos = "From part %s (at position %d)" % (part, index)
                except ValueError:
                    infos = "No containing part could be identified"
                available_primers_names.append(name)
                primer = Primer(name, primer_seq,
                                metadata={'available': False, 'infos': infos})
            selected_primers.append(primer)
            selected_primer_from_seq[primer_seq] = primer

        # CHOOSE A MINIMAL PRIMER COVER FOR EACH CONSTRUCT
        per_record_primers = []
        for record in self.logger.iter_bar(record=records):
            elements = set(indices_to_cover[record.id].values())
            subcovers = {
                prim_seq: primers_coverages[prim_seq].intersection(elements)
                for prim_seq in primers_cover
            }
            # subsets = deepcopy(list(subcovers.items()))
            subsets = list(subcovers.items())
            sub_primers_cover = minimal_cover(elements, subsets=subsets,
                                              heuristic=heuristic)

            # All primers selected for this construct, sorted by availability.
            sub_selected_primers = sorted([
                selected_primer_from_seq[primer_seq]
                for primer_seq in sub_primers_cover
            ], key=lambda p: p.metadata['available'])

            per_record_primers.append(sub_selected_primers)

        return per_record_primers

    def compute_indices_to_cover(self, record):
        """List all indices in the record which should be covered.

        These indices are equidistant points inside the user-defined zones to
        cover in the record.

        The use determines the zones to cover via features of
        type ``misc_feature`` and label 'cover'.
        """
        segments_to_cover = [
            sorted([int(f.location.start), int(f.location.end)])
            for f in record.features
            if f.type == 'misc_feature'
            and "".join(f.qualifiers.get('label', '')) == 'cover'
        ]
        res = self.coverage_resolution
        return set([
            int(indice) % len(record)
            for (start, end) in segments_to_cover
            for indice in np.linspace(start, end, 1.0 * (end - start) / res)
        ])

    @staticmethod
    def locate_primer_sequence(primer, sequence):
        """Find the location (start, end, strand) of a primer in the sequence.

        Return None if the primer sequence and its reverse complement are not
        found in the sequence.
        """
        ind = sequence.find(primer)
        strand = 1
        if ind == -1:
            ind = sequence.find(reverse_complement(primer))
            if ind == -1:
                return None
            else:
                strand = -1
        start, end = ind, ind + len(primer)
        return start, end, strand

    def compute_forbidden_patterns_locations(self, record):
        """Return an array where ``arr[i] == 1`` means that i is surrounded by
        a user-forbidden pattern."""
        pattern_constraints = [AvoidPattern(homopolymer_pattern(c, 5))
                               for c in 'ATGC']
        kmer_constraints = [AvoidPattern(repeated_kmers(k, n))
                            for k, n in [(4, 2), (3, 3), (2, 4)]]
        problem = DnaOptimizationProblem(
            sequence=record,
            constraints=pattern_constraints + kmer_constraints
        )
        constraints_breaches = group_overlapping_segments([
            (f.location.start, f.location.end)
            for f in problem.constraints_evaluations().locations_as_features()
        ])
        return segments_to_array(constraints_breaches, len(record))

    def compute_user_forbidden_locations(self, record):
        """Return an array where ``arr[i] == 1`` means that i is surrounded by
        a user-forbidden location."""
        forbidden_segments = [
            sorted([int(f.location.start), int(f.location.end)])
            for f in record.features
            if f.type == 'misc_feature'
            and "".join(f.qualifiers.get('label', '')) == 'no_primer'
        ]
        return segments_to_array(forbidden_segments, len(record))

    def compute_nonunique_segments_locations(self, record):
        """Return an array where ``arr[i] == 1`` means that i is surrounded by
        a non-unique location."""
        sequence = str(record.seq)
        non_unique_segments = find_non_unique_segments(sequence)
        return segments_to_array(non_unique_segments, len(record))

    def compute_all_forbidden_locations(self, record):
        """Return an array indicating which positions should be avoided.

        We take into account forbidden patterns, user-forbidden locations,
        and non-unique locations.
        ``arr[i] == 1`` indicates that position i should be avoided in the
        record.

        """
        return np.maximum(*(f(record) for f in (
            self.compute_nonunique_segments_locations,
            self.compute_forbidden_patterns_locations,
            self.compute_user_forbidden_locations
        )))

    def compute_sequence_primers(self, record):
        """Return, primers for the sequence, one around each index.

        The primers are chosen to fit the length and melting temperature
        specified by the class parameters.

        Parameters
        ----------
        record
          The record in which to list the primers

        Returns
        -------
        primers
          List of primers sequences.

        """
        sequence = str(record.seq)
        L = len(sequence)
        rev_sequence = reverse_complement(sequence)
        locations = find_best_primer_locations(
            sequence,
            tm_range=self.tm_range,
            size_range=self.size_range
        )
        return list(set(sum([
            [
                (sequence[l[0]: l[1]], (l[0], l[1], 1)),
                (rev_sequence[L - l[1]: L - l[0]], (l[0], l[1], -1))
            ]
            for l in locations
            if l is not None
        ], [])))

    def compute_all_primers_coverage_on_record(self, record, indices_to_cover,
                                               available_primers):
        """Return, for each primer, the list of indices covered in this record.

        Parameters
        ----------
        record
          The record in which to search

        indices_to_cover
          List of indices to cover (defined by the user) in this record.

        available_primers
          List of candidate primers.

        Returns
        -------
        primers_coverage
          ``{primer_name: list_of_indices_covered_in_this_record}``.

        """
        sequence = str(record.seq)
        linear = record.__dict__.get('linear', True)
        L = len(sequence)
        forbidden_locations = self.compute_all_forbidden_locations(record)
        sequence_primers = self.compute_sequence_primers(record)
        reusable_primers_sequences = [
            (primer, location) for (primer, location) in [
                (primer, self.locate_primer_sequence(primer, sequence))
                for primer in available_primers
            ]
            if location is not None
        ]
        primers = reusable_primers_sequences + sequence_primers
        iter_bar = self.logger.iter_bar
        primers_coverages = {}
        for primer, (start, end, strand) in iter_bar(primer=primers):
            if forbidden_locations[start:end].max():
                continue
            if not all(cond(primer) for cond in self.primers_conditions):
                continue

            if strand == 1:
                coverage_start = end + self.read_range[0]
                coverage_end = end + self.read_range[1]
            else:
                coverage_start = start - self.read_range[1]
                coverage_end = start - self.read_range[0]
            if (not linear) and (coverage_start < 0):
                a, b, c, d = -1000, coverage_end, L + coverage_start, L + 1000
            elif (not linear) and (coverage_end > L):
                a, b, c, d = -1000, coverage_end - L, coverage_start, L + 1000
            else:
                a, b, c, d = (coverage_start, coverage_end,
                              coverage_start, coverage_end)
            primers_coverages[primer] = set([
                indices_to_cover[ind]
                for ind in indices_to_cover
                if (a <= ind <= b) or (c <= ind <= d)
            ])
        return primers_coverages

    def generate_primer_name(self, prefix='P', available_primers_names=(),
                             n_digits=6):
        """Return a suitable primer name, considering existing primer names.

        The result will be of the form P000425 where 'P' is the prefix and
        '000425' means that 'P000424' was the highest-numbered primer name
        sarting with P in the list of available primer names

        Parameters
        ----------
        prefix
          The prefix for the primers name

        available_primers_names
          List of already-allocated primer names
        """
        max_index = 0
        for name in available_primers_names:
            if not name.startswith(prefix):
                continue
            index = int(re.search(r'\d+', name[len(prefix):]).group())
            if (index is not None) and (index > max_index):
                max_index = index
        return prefix + str(max_index + 1).zfill(n_digits)

    def find_part_name_in_record(self, record, index):
        """Find a part where the sequence appears in the provided record.

        This is used to provide primers infos ("where does this come from ?").

        Parameters
        ----------
        sequence
          An ATGC string representing a sequence (of a primer)

        record
          A single Biopython records where to find the sequence. The parts
          should be marked by features with a qualifier ``{'part': part_name}``

        Returns
        -------
        part_name, index
          The name of the part, and position in the part, where it was found.
          If the sequence is found in different parts, only the first find
          is returned.
        """
        for f in record.features:
            if 'part' not in str(f.qualifiers):
                continue
            part_name = ''.join(f.qualifiers['part'])
            strand = f.location.strand
            start, end = sorted([int(e) for e in [f.location.start,
                                                  f.location.end]])
            if start <= index <= end:
                distance = (index - start) if (strand >= 0) else (end - index)
                return (part_name, distance)
        else:
            raise ValueError(
                'The given index %d does not correspond to any part in "%s"'
                % (index, record.id)
            )

    def find_subsequence_in_records(self, sequence, records):
        """Find a part where the sequence appears in the provided records.

        This is used to provide primers infos ("where does this come from ?").
        This will look for either the sequence or its reverse-complement.

        Parameters
        ----------
        sequence
          An ATGC string representing a sequence (of a primer)

        records
          A list of Biopython records where to find the sequence. The parts
          should be marked by features with a qualifier ``{'part': part_name}``

        Returns
        -------
        part_name, index
          The name of the part, and position in the part, where it was found.
          If the sequence is found in different parts, only the first find
          is returned.
        """
        for r in records:
            ind = r.seq.find(sequence)
            if ind > 0:
                part_name, index = self.find_part_name_in_record(r, ind)
                break
            else:
                ind = r.seq.reverse_complement().find(sequence)
                if ind > 0:
                    part_name, index = self.find_part_name_in_record(r, ind)
                    break
        else:
            raise ValueError('sequence not found in records')
        return (part_name, index)

    def name_subsequence_in_records(self, sequence, records, prefix='P'):
        """Write a table of primers with columns 'name', 'sequence', etc.

        The columns after 'sequence' are one column per primer metadata, such
        as 'available', 'infos', etc.

        Parameters
        ----------
        primers
          A list of primers, or list of list, as returned by ``select_primers``

        csv_path
          The path to a csv file to write to. If None, no file is written, only
          the pandas dataframe of the table is returned.

        Returns
        -------
        dataframe
          The pandas dataframe of the table.
        """
        for r in records:
            ind = r.seq.find(sequence)
            if ind > 0:
                part_name, index = self.find_part_name_in_record(r, ind)
                break
            else:
                ind = r.seq.reverse_complement().find(sequence)
                if ind > 0:
                    part_name, index = self.find_part_name_in_record(r, ind)
                    break
        else:
            raise ValueError('sequence not found in records')
        return "%s%s_%04d" % (prefix, part_name, index)

    def plot_coverage(self, records, selected_primers, pdf_path):
        """Plot the predicted sequencing coverage for each construct."""
        # PLOT THE PREDICTED SEQUENCING COVERAGE FOR EACH CONSTRUCT

        with PdfPages(pdf_path) as pdf:
            iterator = zip(records, selected_primers)
            self.logger(message='Plotting coverages...')
            for record, primers in self.logger.iter_bar(record=iterator):
                matches_set = simulate_sequencing(
                    record, primers=primers, read_range=self.read_range,
                    linear=record.__dict__.get('linear', True)
                )
                ax = matches_set.plot(plot_reference=True, plot_coverage=False)
                ax.set_title(record.id, fontsize=14, weight='bold')
                for x in self.compute_indices_to_cover(record):
                    ax.axvline(x, c='#fceddb', lw=2, ls='--', zorder=-2000)
                pdf.savefig(ax.figure, bbox_inches='tight')
                plt.close(ax.figure)

    def write_records_primers_table(self, selected_primers, records, sep='|',
                                    csv_path=None):
        """Write a table with columns 'construct','primers' for this construct.

        Parameters
        ----------
        selected_primers
          The list of list of primers, as returned by the ``select_primers``
          method.

        records
          The list of records, as provided to the ``select_primers`` method.

        sep
          The separator between the different primer names in the ``primers``
          column. Avoid ';' or ',' as this might be used by the CSV formatter.

        csv_path
          The path to a csv file to write to. If None, no file is written, only
          the pandas dataframe of the table is returned.

        Returns
        -------
        dataframe
          The pandas dataframe of the table.
        """
        df = pandas.DataFrame.from_records([
            {
                'record': record.id,
                'primers': sep.join([p.name for p in primers_list])
            }
            for record, primers_list in zip(records, selected_primers)
        ], columns=['record', 'primers'])
        if csv_path is not None:
            df.to_csv(csv_path, index=False)
        return df

    def write_primers_table(self, selected_primers, csv_path=None):
        """Write a table of primers with columns 'name', 'sequence', etc.

        The columns after 'sequence' are one column per primer metadata, such
        as 'available', 'infos', etc.

        Parameters
        ----------
        primers
          A list of primers, or list of list, as returned by ``select_primers``

        csv_path
          The path to a csv file to write to. If None, no file is written, only
          the pandas dataframe of the table is returned.

        Returns
        -------
        dataframe
          The pandas dataframe of the table.
        """
        if isinstance(selected_primers[0], (list, tuple)):
            selected_primers = set([p for pp in selected_primers for p in pp])
        df = Primer.list_to_spreadsheet(selected_primers)
        df = df.sort_values(by=['available', 'name'])
        if csv_path is not None:
            df.to_csv(csv_path, index=False)
        return df

    def write_multifile_report(self, records, selected_primers,
                               target='@memory'):
        root = flametree.file_tree(target)
        self.plot_coverage(
            records=records,
            selected_primers=selected_primers,
            pdf_path=root._file('coverages_plots.pdf').open('wb'))
        self.write_primers_table(
            selected_primers=selected_primers,
            csv_path=root._file('primers_list.csv')
        )
        self.write_records_primers_table(
            records=records,
            selected_primers=selected_primers,
            csv_path=root._file('primers_per_record.csv')
        )

        return root._close()
